////////////////////////////////////////////////////////////
// btreet1.cpp: Btree "regression test" program.
// Generates random keys, inserts them, deletes them, and
// checks for consistency along the way.
// Copyright(c) 1993 Azarona Software. All rights reserved.
/////////////////////////////////////////////////////////////
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include "btree.h"

void Pause(int on_auto)
{
  if (on_auto) return;
  printf("\n\nPress return to continue");
  getchar(); putch('\n'); putch('\r');
}

void Stats(Btree &t)
{
  printf("\n**** STATISTICS ****\n");
  t.Statistics(0);
}

const int MAXNKEYS = 1024;

char keys[MAXNKEYS][MAXKEYSIZE];
char exists[MAXNKEYS];

int NKEYS = 512;

main(int argc, char *argv[])
{
  Entry e;
  int rv;

  int verbose = 1;
  int on_auto = 1;
  int csize = 8;

  unsigned seed = (unsigned)time(NULL);

  int i = 1;
  while(i < argc) {
    if (!strcmp(argv[i], "-p")) {
       on_auto = 0;
    }
    else if (!strcmp(argv[i], "-n")) {
       NKEYS = atoi(argv[++i]);
       if (NKEYS > MAXNKEYS) NKEYS + MAXNKEYS;
    }
    else if (!strcmp(argv[i], "-q")) {
       verbose = 0;
    }
    else if (!strcmp(argv[i], "-c")) {
       csize = atoi(argv[++i]);
    }
    else {
       seed = (unsigned)atoi(argv[i]);
    }
    i++;
  }

  srand(seed);

  printf("Seed is: %u\n", seed);
  printf("Making index file with cache size %d\n", csize);

  FmgrPtr fptr(new Fmgr);
  Btree idx1(csize);

  fptr->Create("test.idx");
  idx1.Connect(fptr, 1);
  printf("Indexing %d items:\n", NKEYS);

  Pause(on_auto);

  printf("First pass: Adding %d random entries\n", NKEYS);

  // add NKEYS keys to each index file
  for (i = 0; i < NKEYS; i++) {
      if (verbose) printf("%d ", i);
      int j, rlen;
      rlen = random(MAXKEYSIZE-1) + 1;
      for (j=0; rlen > 0; rlen--, j++) {
          keys[i][j] = random(26) + 'A';
      }
      keys[i][j] = 0;
      rv = idx1.Add(keys[i], i);
      if (rv != SUCCESS) {
         printf("\nProblem adding '%s' - %d\n", keys[i], i);
         excp->THROW(ASSERTERR);
      }
      exists[i] = 1;
  }

  printf("\nIndexing is complete\n");

  if (verbose) Stats(idx1);

  Pause(on_auto);

  printf("Verifying insertions\n");

  for (i = 0; i<NKEYS; i++) {
      if (verbose) printf("%d ", i);
      e = keys[i];
      e.data = i;
      rv = idx1.FullSearch(e);
      if (rv != SUCCESS && exists[i]) {
         printf("\nProblem finding '%s' - %d\n", keys[i], i);
         excp->THROW(ASSERTERR);
      }
  }

  Pause(on_auto);

  printf("\nDeleting random entries\n");

  for (i = 0; i < NKEYS; i++) {
      int r = random(10);
      if (r == 1) {
         if (verbose) printf("%d ", i);
         if (idx1.Remove(keys[i], i) != SUCCESS) {
            printf("\nProblem removing '%s' - %d\n", keys[i], i);
            excp->THROW(ASSERTERR);
         }
         exists[i] = 0;
      }
  }
  if (verbose) printf("\n");

  if (verbose) Stats(idx1);

  Pause(on_auto);

  printf("\nChecking static consistency\n");

  for(i = 0; i<NKEYS; i++) {
     if (verbose) printf("%d ", i);
     e = keys[i];
     e.data = i;
     rv = idx1.FullSearch(e);
     if (rv == SUCCESS) {
        if (verbose) printf("found     ");
        if (exists[i]) {
           if (verbose) printf("OK");
        }
        else {
          printf("\nFound '%s' - %d, but shouldn't have\n", keys[i], i);
          excp->THROW(ASSERTERR);
        }
     }
     else {
        if (verbose) printf("not found ");
        if (exists[i]) {
           printf("\nShould have found '%s' - %d, but couldn't\n", keys[i], i);
           excp->THROW(ASSERTERR);
        }
        else if (verbose) printf("OK");
     }
     if (verbose) printf("\n");
  }

  Pause(on_auto);

  printf("\nAdding deleted entries back\n");

  for (i = 0; i<NKEYS; i++) {
      if (!exists[i]) {
         if (verbose) printf("%d ", i);
         rv = idx1.Add(keys[i], i); 
         if (rv != SUCCESS) {
            printf("\nProblem adding '%s' - %d\n", keys[i], i);
            excp->THROW(ASSERTERR);
         }
         exists[i] = 1;
      }
  }

  Pause(on_auto);

  printf("\nVerifying insertions again\n");

  for (i = 0; i<NKEYS; i++) {
      if (verbose) printf("%d ", i);
      e = keys[i];
      e.data = i;
      rv = idx1.FullSearch(e);
      if (rv != SUCCESS && exists[i]) {
         printf("\nProblem finding '%s' - %d\n", keys[i], i);
         excp->THROW(ASSERTERR);
      }
  }


  Pause(on_auto);

  printf("Checking dynamic consistency by deleting all entries\n");

  for(i = 0; i<NKEYS; i++) {
     if (verbose) printf("%d ", i);
     rv = idx1.Remove(keys[i], i);
     if (rv == SUCCESS) {
        if (verbose) printf("deleted     ");
        if (exists[i]) {
           if (verbose) printf("OK");
        }
        else {
          printf("\nFound '%s' - %d to delete, but shouldn't have\n", keys[i], i);
          excp->THROW(ASSERTERR);
        }
     }
     else {
        if (verbose) printf("not found   ");
        if (exists[i]) {
           printf("\nCouldn't find '%s' - %d to delete, but should have\n", keys[i], i);
           excp->THROW(ASSERTERR);
        }
        else if (verbose) printf("OK");
     }
     if (verbose) printf("\n");
  }

  if (verbose) Stats(idx1);

  printf("\n\n---> Test Succeeded <---\n");
  printf(    "     Flusing buffers ...\n");
  return 0;
}
