////////////////////////////////////////////////////////////////////
// coptrb.cpp: Cached object pointer base class methods.
// Does all the things that don't depend on the data stored.
// Copyright(c) 1993 Azarona Software. All rights reserved.
////////////////////////////////////////////////////////////////////
#include "coptrb.h"
#include "bucketb.h"
#include "cacheb.h"

Coptrb::Coptrb(Cacheb &c, long p)
// General coptr constructor. Note that a cache
// is required. The coptr stays unbound till needed.
{
  cache = &c; addr = p; bound = 0; bkt = 0;
}

Coptrb::Coptrb(const Coptrb &c)
// Copy constructor. Note that destination coptr
// won't be bound until needed, even if source is bound.
{
  addr = c.addr; bkt = c.bkt; cache = c.cache; bound = 0;
}

Bucketb *Coptrb::Alloc(size_t n, long p)
// Allocates n bytes of room (presumably the size of the
// data in the bucket) at address p in the cache's
// file, and sets up a cache bucket to support it.
// If p != 0, it means the room has already been allocated.
// Note that ReserveBkt will setup bucket's address,
// cache pointer and refcnt, but doesn't load any data.
// Returns a pointer to the bucket, or 0 if couldn't
// allocate.
{
  Release(); // Let go of any currently bound data
  if (p == 0) p = cache->fptr->Alloc(n);
  bkt = cache->ReserveBkt(p, 0);
  if (bkt) {
     bkt->SetDirty(); // Data modified by constructor
     addr = p;
     bound = 1;
  }
  else {
     addr = 0; // bound will = 0 here as well
  }
  return bkt;
}

void Coptrb::Delete(unsigned n)
// Deallocates the file data pointed to by this coptr, supposedly
// of size n. Deallocation only takes place if this coptr is
// pointing to a bucket bound by no one else. Coptr is unbound
// afterwards and both the bucket and Coptr are made null.
{
  if (bound && bkt->IsLocked() == 1 ||
      !bound && (!bkt || !bkt->IsLocked())) {
     cache->fptr->Free(n, addr);
     Release();
     bkt->MakeNull();
     addr = 0;
  }
  else excp->THROW(DANGLINGPTR);
}

void Coptrb::Copy(const Coptrb &c)
// Copies one coptr into another. Note that the destination
// coptr isn't bound, even if the source coptr is.
{
  Release();  
  addr = c.addr;
  bkt = c.bkt;
  bound = 0;
  cache = c.cache;
}

void Coptrb::Bind()
// Binds the coptr to a bucket containing data stored at
// file location addr. Note that we may already be pointing
// to the correct bucket, so check for that.
// ASSUMES not already bound. For internal use only.
{
  if (bkt && bkt->addr == addr) { // Fast binding
     if (addr) {
        bkt->Lock();
        cache->FastBind();
        bound = 1;
     }
     else excp->THROW(NULLPTR);
  }
  else {
     bkt = cache->ReserveBkt(addr, 1);
     if (bkt) bound = 1; else excp->THROW(NULLPTR);
  }
}

void Coptrb::Release()
// Signal to the bucket that we're through with it.
{
  if (bound) {
     bkt->Unlock();
     bound = 0;
  }
}

void Coptrb::NotifyUseOf()
// Moves the referenced bucket to the front of the
// cache queue. For cache performance tuning.
{
  if (bound) cache->MoveToFront(bkt);
}
