///////////////////////////////////////////////////////////////////
// fmgr.cpp: File mananger class methods.
// Copyright (c) 1989-1993 Azarona Software. All rights reserved.
///////////////////////////////////////////////////////////////////
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include "fmgr.h"

// Allocate all necessary static data

static char closed_file_name[] = "closed-file";
char Fmgr::sg[4] = {'F', 'M', 'G', 'R'};

/////////////////////////////////////////////
// FmgrPtr methods
/////////////////////////////////////////////

void FmgrPtr::Bind(const FmgrPtr &p)
// Binds object to the same object that p is bound to.
// Assumes we're not already bound to an object.
{
  objptr = p.objptr;
  if (objptr) objptr->IncRef();
}

void FmgrPtr::Unbind()
// Unbinds from the object we're bound to. Assumes
// we really are bound to an object. Also, if
// the reference count goes to zero for the object,
// it better really be stored on the heap.
{
  if (objptr == 0) return;
  objptr->DecRef();
  if (objptr->UnReferenced()) delete objptr;
}

void FmgrPtr::NewBinding(const FmgrPtr &p)
// Gives the smart pointer a new binding--the same as p has.
// Note that the test for using the same binding as before
// isn't there on a whim. It really is important in preventing
// accidental deletions, in case we try to bind to ourself,
// and we're the only reference to the object we're bound to.
{
  if (objptr != p.objptr) {
     Unbind();
     Bind(p);
  }
}

////////////////////////////////////////////////////
// Fmgr methods
////////////////////////////////////////////////////

void Fmgr::Throw(ExceptionCode ec, char *src, int lno, char *msg)
// We fix up some of the exception codes based on context.
{
  if (ec == NOTWRITEABLE && OK()) ec = READONLYERR;
  if (ec == NOTREADY && (status & 1)) ec = NOTOPENERR;
  excp->Throw(ec, src, lno, msg, name);
  status &= 0xfe; // Reset good bit
}

Fmgr::Fmgr()
// Creates a file manager object. Note that an implicit call 
// to the base case constructor CountObj() will set the 
// reference count for this object to one.
{
  strcpy(name, closed_file_name);
  fp = 0; status = 0x01; // good, read-only, and closed
}

Fmgr::~Fmgr()
// Destructor for file object.
{
  Close();
}

int Fmgr::Create(char *fname, long static_sz)
// Creates and opens a new file named fname,
// truncating it if it already exists. The
// area at the front of the file of length
// static_sz + sizeof(FmgrHeader) is reserved.
// Returns 1 if the file was successfully created
// and opened, else 0.
{

  // First, close the current file if open.
  Close();

  status = 0x05; // Set read/write bit and good bit

  // Create, truncate if already exists
  fp = ::fopen(fname, "w+b");

  if (fp == 0) {
     THROW(CREATERR); // Not intended to terminate program
  }
  else {
    status |= 2; // Set open bit
    strcpy(name, fname);
    h.hs = static_sz + sizeof(FmgrHeader);
    lastop = fetch; // So Store() works right the first time
    InitHdr();
  }

  return IsOpen();
}

void Fmgr::InitHdr()
// Sets up the header area in the file. If there's
// statically allocated data beyond the header, write
// a 0 to the last byte of it, so that you won't get
// unexpected end of file errors. Note that the rest
// of the static area stays uninitialized.
{
  h.fs = 0;
  h.fe = h.hs;
  memcpy(h.sg, sg, 4); // Copy signatures
  WriteHdr();
  if (h.hs > sizeof(FmgrHeader)) {
     char zero_byte = 0;
     Store(&zero_byte, 1, h.hs-1);
  }
}

int Fmgr::Open(char *fname, AccessMode mode)
// Opens the fname file. File must exist and be an Fmgr-type 
// file or error occurs. First closes the current file if open.
// Returns 1 if file opened successfully, else 0.
{
  char *mode_str;

  // First, close the current file if open.
  Close();

  if (mode == read_only) {
     mode_str = "rb";
     status = 0x01; // set good bit, reset read/write bit
  }
  else {
     mode_str = "r+b";
     status = 0x05; // Set good bit and read/write bit
  }

  fp = ::fopen(fname, mode_str);

  if (fp == 0) {
     THROW(OPENERR); // Not intended to terminate program
  }
  else {
    status |= 2; // Set open bit
    strcpy(name, fname);
    lastop = store; // So Fetch() works right the first time
    ReadHdr();
    if (memcmp(h.sg, sg, 4)) { // Test file type
       Close(0);
       THROW(FTYPERR); // Not intended to terminate program
    }
  }

  return IsOpen();
}

void Fmgr::Close(int flush)
// Closes the file if not already closed, flushing the
// basic header if flush = 1. Does nothing if in the
// error state. Checks for dangling references to
// this file.
{
  if (OK()) {
     if (refcnt > 1) THROW(DANGLINGPTR);
     if (flush && !ReadOnly())
        WriteHdr();
     if (::fclose(fp) != 0) THROW(CLOSERR);
     strcpy(name, closed_file_name);
  }
  status &= 0xfd; // Reset open bit
}

void Fmgr::Flush()
// Writes the Fmgr header to the file, and then flushes
// any internal buffers the file might have.
{
  if (ReadyForWriting()) {
     WriteHdr();
     if (fflush(fp) != 0) THROW(RWERR);
     Seek(0, SEEK_CUR); // So that my lastop logic works right
  }
}


void Fmgr::Seek(long ofs, int seek_mode)
// Moves the file pointer to the byte offset ofs,
// using seek_mode, (which should be either SEEK_SET,
// SEEK_CUR, or SEEK_END).
{
  if (OK()) {
     int rv = ::fseek(fp, ofs, seek_mode);
     if (rv != 0) THROW(SEEKERR);
     lastop = seek;
  }
  else THROW(NOTREADY);
}

void Fmgr::Fetch(void *d, unsigned nbytes, long p)
// Fetches nbytes from address p into buffer d. The address 
// is always interpreted to be from the beginning of the
// file, unless it's CURRADDR, which means from the
// current position.
{
  if (OK()) {
     // Ensure seeks between intervening reads and writes,
     // and optimize for sequential reads
     if (p == CURRADDR) {
        if (lastop == store) Seek(0, SEEK_CUR);
     }
     else Seek(p, SEEK_SET);
     unsigned bytesmoved = ::fread(d, 1, nbytes, fp);
     if (bytesmoved != nbytes) THROW(feof(fp) ? EOFERR : RWERR);
     lastop = fetch;
  }
  else THROW(NOTREADY);
}

void Fmgr::Store(const void *d, unsigned nbytes, long p)
// Stores nbytes from buffer d to addr p. The addr is
// always interpreted to be from the beginning of the
// file, unless it's CURRADDR, which means from the
// current position.
{
  if (ReadyForWriting()) {
     // Ensure seeks between intervening reads and writes,
     // and optimize for sequential writes
     if (p == CURRADDR) {
        if (lastop == fetch) Seek(0, SEEK_CUR);
     }
     else Seek(p, SEEK_SET);
     unsigned bytesmoved = ::fwrite(d, 1, nbytes, fp);
     if (bytesmoved != nbytes) THROW(feof(fp) ? EOFERR : RWERR);
     lastop = store;
  }
  else THROW(NOTWRITEABLE);
}

void Fmgr::ReadHdr()
// Reads the file header.
{
  Fetch(&h, sizeof(FmgrHeader), 0L);
}

void Fmgr::WriteHdr()
// Writes the file header.
{
  Store(&h, sizeof(FmgrHeader), 0L);
}

void Fmgr::FetchFBlkHdr(FBlkHeader &h, long p)
// Reads in free block header, and tests check word
// to make sure the file is still in sync.
{
  Fetch(&h, sizeof(FBlkHeader), p);
  if (h.check_word != FREEBLK) THROW(CORRUPTERR);
}

long Fmgr::Reclaim(unsigned nbytes)
// Finds the first block on the free space list that is big
// enough for nbytes of data. Puts back on the free list
// all those bytes that aren't needed. Note that the amount
// put back must be at least the size of the free block 
// header, otherwise, the entire block is considered not
// an appropriate size and rejected. Returns address of the 
// newly reclaimed data, or 0 if there wasn't a free space 
// block that was appropriate.
{
  FBlkHeader bh, prev_bh, new_bh;
  long addr, prev_addr, new_addr;
  unsigned avail_len, unused_len;

  addr = h.fs; prev_addr = 0;

  while(addr) { // Until a block of a valid size is found
    FetchFBlkHdr(bh, addr);
    if (!OK()) break;
    avail_len = bh.len + sizeof(FBlkHeader);
    if (avail_len >= nbytes + sizeof(FBlkHeader))
       unused_len = avail_len - nbytes;
       else unused_len = 0;
    if (avail_len == nbytes) {
       // Block is an exact fit, so link prev link to next link
       if (prev_addr == 0) {
          // We're at the head of freespace, so we have a new head
          h.fs = bh.next;
       }
       else {
          // In the middle of free space, so link prev to next
          prev_bh.next = bh.next;
          StoreFBlkHdr(prev_bh, prev_addr);
       }
       break;
    }
    else if (unused_len > 0) {
       // Block too big, and there's room for a free block
       // header in the unused portion. So splice in this
       // new block.
       new_addr = addr + nbytes;
       new_bh.check_word = FREEBLK;
       new_bh.next = bh.next;
       new_bh.len = unused_len - sizeof(FBlkHeader);
       StoreFBlkHdr(new_bh, new_addr);
       if (prev_addr == 0) {
          // We're at the head of freespace, so we have a new head
          h.fs = new_addr;
       }
       else { // In the middle of freespace
          prev_bh.next = new_addr;
          StoreFBlkHdr(prev_bh, prev_addr);
       }
       break;
    }
    else {
      // Block not big enough, so try next block
      prev_addr = addr;
      prev_bh = bh;
      addr = bh.next;
    }
  } // End of looking for big enough block

  return OK() ? addr : 0;
}


long Fmgr::Alloc(unsigned nbytes)
// Allocates a block of nbytes of data, either from the free space list,
// or from the end of the file. The minimum number of bytes allocated
// is equal to sizeof(BlkHeader). Nothing is written to the newly 
// allocated space.
// Returns location of space allocated, or returns a 0 if some
// error occurred.
{
  long p = 0;

  if (ReadyForWriting()) {

     // Adjust number of bytes to allocate to minimum size

     if (nbytes < sizeof(FBlkHeader)) nbytes = sizeof(FBlkHeader);

     // Try using a free block
     p = Reclaim(nbytes);

     if (OK() && p == 0) { // We may have to extend the file instead
        p = h.fe;
        h.fe += nbytes;
        // Write to the last byte to avoid possibe end of file errors
        char zero = 0;
        Store(&zero, sizeof(char), h.fe-1);
     }
  }
  else THROW(NOTWRITEABLE);
  return p;
}

void Fmgr::Free(unsigned nbytes, long p)
// Frees up the block at location p ASSUMED to be
// nbytes in size. Block is placed on the front
// of the free space list.
// If nbytes is < sizeof(FBlkHeader), it is forced
// to that size, since that's the minimum size
// allocated.
{
  FBlkHeader fh;

  if (ReadyForWriting()) {
     if (nbytes < sizeof(FBlkHeader)) nbytes = sizeof(FBlkHeader);
     fh.check_word = FREEBLK;
     fh.len = nbytes - sizeof(FBlkHeader);
     // Block to become head of free list
     fh.next = h.fs;
     StoreFBlkHdr(fh, p);
     h.fs = p;
  }
  else THROW(NOTWRITEABLE);
}
