/////////////////////////////////////////////////////////////
// Mnode.cpp: Multi-way node methods
// Copyright(c) 1993 Azarona Software. All rights reserved.
/////////////////////////////////////////////////////////////
#include <string.h>
#include "mnode.h"

Entry::Entry(const Entry &src)
{
  strcpy(key, src.key);
  data = src.data;
  right = src.right;
}

Entry::Entry(const char *s)
{
  strncpy(key, s, MAXKEYSIZE); key[MAXKEYSIZE-1] = 0;
  data = 0; right = 0;
}

Entry::Entry(const char *s, long d)
{
  strncpy(key, s, MAXKEYSIZE); key[MAXKEYSIZE-1] = 0;
  data = d; right = 0;
}

void Entry::operator=(const Entry &src)
{
  if (this == &src) return;
  strcpy(key, src.key);
  data = src.data;
  right = src.right;
}

void Entry::operator=(const char *s)
// Assigns the key portion, but no other fields
{
  strncpy(key, s, MAXKEYSIZE); key[MAXKEYSIZE-1] = 0;
}

int Compare(const Entry &a, const Entry &b)
// Friend function that does a partial compare of a and b.
// Compares only the key string, not the data.
// Returns -1 if a < b, 0 if a == b, or 1 if a > b.
{
   return strcmp(a.key, b.key);
}

int FullCompare(const Entry &a, const Entry &b)
// Friend function that does a full compare of a and b.
// Compares the key string as well as the data.
// Returns -1 if a < b, 0 if a == b, or 1 if a > b.
{
   int rv = strcmp(a.key, b.key);
   if (rv > 0) return 1;
   if (rv < 0) return -1;
   if (a.data > b.data) return 1;
   if (a.data < b.data) return -1;
   return 0;
}

int Mnode::Search(const Entry &e, int &posn)
// Tries to match e's key with the keys in the entries
// of this node. Returns -1 if e's key is less than all
// keys in the node, returns 0 if there was a match, 
// return 1 if e's key is greater than all keys in the 
// node. Passes back posn of matching entry if any, or 
// the posn of the entry containing the appropriate branch
// to keep searching down.
{
  posn = cnt-1;

  while(posn >= 0) {
    int rv = Compare(e, entry[posn]);
    if (rv > 0) return 1;
    if (rv == 0) return 0;
    posn--;
  }
  return -1;
}

int Mnode::FullSearch(const Entry &e, int &posn)
// Like the first Search(), except we use FullCompare().
{
  posn = cnt-1;

  while(posn >= 0) {
    int rv = FullCompare(e, entry[posn]);
    if (rv > 0) return 1;
    if (rv == 0) return 0;
    posn--;
  }
  return -1;
}

void Mnode::Split(Mnode &b, int split_posn)
// Moves right half of this node at split_posn into empty b.
// Assumes split_posn is in range.
{
  b.cnt = cnt - split_posn; // Compute how much to move
  memmove(b.entry, entry+split_posn, (b.cnt)*sizeof(Entry));
  cnt = split_posn;
}

void Mnode::InsEntry(Entry &e, int posn)
// Inserts entry e into node at position posn. Assumes
// there is room and assumes posn <= cnt;
{
  memmove(entry+posn+1, entry+posn, (cnt-posn)*sizeof(Entry));
  entry[posn] = e;
  cnt++;
}

void Mnode::Concatenate(Mnode &n)
// Adds all entries of node n to the end of this node.
// Assumes there is room.
{
  for(int i = 0; i<n.cnt; i++) entry[cnt++] = n.entry[i];
}

void Mnode::DelEntry(int posn)
// Deletes the entry at position posn. Assumes posn is in range.
{
  cnt--;
  memmove(entry+posn, entry+posn+1, (cnt-posn)*sizeof(Entry));
}
