/////////////////////////////////////////////////////////////
// mnode.h: Multi-way node class definition.
// Copyright(c) 1993 Azarona Software. All rights reserved.
/////////////////////////////////////////////////////////////
#ifndef H_MNODE
#define H_MNODE

// For maximum testing, using a small order, such as 3. This
// will give the insert/delete routines a more thorough workout.
// But for real use, you'll want a higher order. Too high, however,
// and the code will spend too much time reading in large nodes.

const int MAXKEYSIZE = 16; // Maximum size of a key
const int NBR = 3;         // Order of the tree

struct Entry {
  char key[MAXKEYSIZE]; 
  long data;
  long right; // Points to right child
  Entry();
  Entry(const char *s);
  Entry(const char *s, long d);
  Entry(const Entry &src);
  void operator=(const Entry &src);
  void operator=(const char *s);
  friend int Compare(const Entry &a, const Entry &b);
  friend int FullCompare(const Entry &a, const Entry &b);
};

inline Entry::Entry()
// Leave everything uninitialized, for sake of speed
{
}

struct Mnode {
  int cnt;   // Must come before left
  long left; // Points to leftmost child. Must come before entry[].
  Entry entry[NBR-1];
  Mnode();
  int Search(const Entry &e, int &posn);
  int FullSearch(const Entry &e, int &posn);
  void Split(Mnode &b, int split_posn);
  void InsEntry(Entry &e, int posn);
  void Concatenate(Entry &e);
  void Concatenate(Mnode &n);
  void DelEntry(int posn);
  long &Branch(int posn);
  int LastPosn();
  int IsEmpty();
  int IsFull();
  int IsPoor();
  int IsPlentiful();
};

inline Mnode::Mnode()
// Set up an empty node
{
  cnt = 0; left = 0;
}

inline long &Mnode::Branch(int posn)
// Returns the branch for the given position. Due to the layout
// of the node, we'll get the left branch if posn = -1.
{
  return entry[posn].right;
}

inline int Mnode::IsEmpty()
{
  return cnt == 0;
}

inline int Mnode::IsFull()
{
  return cnt == NBR-1;
}

inline int Mnode::IsPoor()
// Returns 1 if node has fewer than the minimum entries.
{
  return cnt < (NBR-1)/2;
}  

inline int Mnode::IsPlentiful()
// Returns 1 if node has more than the minimum number of entries.
{
  return cnt > (NBR-1)/2;
}  

inline int Mnode::LastPosn()
// Returns the position of the last entry in the node
{
  return cnt-1;
}

inline void Mnode::Concatenate(Entry &e)
// Adds entry e to the end of the node
{
  InsEntry(e, cnt);
}

#endif
